import collections

import sqlalchemy as sa
from sqlalchemy.exc import InvalidRequestError

from .exceptions import GinoException


class ColumnAttribute:
    def __init__(self, prop_name, column):
        self.prop_name = prop_name
        self.column = column

    def __get__(self, instance, owner):
        if instance is None:
            return self.column
        else:
            return instance.__values__.get(self.prop_name)

    def __set__(self, instance, value):
        instance.__values__[self.prop_name] = value

    def __delete__(self, instance):
        raise AttributeError("Cannot delete value.")


class InvertDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._inverted_dict = dict()
        for k, v in self.items():
            if v in self._inverted_dict:
                raise GinoException(
                    "Column name {} already maps to {}".format(
                        v, self._inverted_dict[v]
                    )
                )
            self._inverted_dict[v] = k

    def __setitem__(self, key, value):
        if value in self._inverted_dict and self._inverted_dict[value] != key:
            raise GinoException(
                "Column name {} already maps to {}".format(
                    value, self._inverted_dict[value]
                )
            )
        super().__setitem__(key, value)
        self._inverted_dict[value] = key

    def invert_get(self, key, default=None):
        return self._inverted_dict.get(key, default)


class Dict(collections.OrderedDict):
    def __setitem__(self, key, value):
        if isinstance(value, sa.Column) and not value.name:
            value.name = key
        return super().__setitem__(key, value)


class ModelType(type):
    def _check_abstract(self):
        if self.__table__ is None:
            raise TypeError(
                "GINO model {} is abstract, no table is "
                "defined.".format(self.__name__)
            )

    def __iter__(self):
        self._check_abstract()
        # noinspection PyUnresolvedReferences
        return iter(self.__table__.columns)

    def __getattr__(self, item):
        try:
            if item in {"insert", "join", "outerjoin", "gino"}:
                self._check_abstract()
                return getattr(self.__table__, item)
            raise AttributeError
        except AttributeError:
            raise AttributeError(
                "type object '{}' has no attribute '{}'".format(self.__name__, item)
            )

    @classmethod
    def __prepare__(mcs, name, bases, **kwargs):
        return Dict()

    def __new__(mcs, name, bases, namespace, **kwargs):
        rv = type.__new__(mcs, name, bases, namespace)
        rv.__namespace__ = namespace
        if rv.__table__ is None:
            rv.__table__ = getattr(rv, "_init_table")(rv)
        return rv


def declared_attr(m):
    """
    Mark a class-level method as a factory of attribute.

    This is intended to be used as decorators on class-level methods of a
    :class:`~Model` class. When initializing the class as well as its
    subclasses, the decorated factory method will be called for each class, the
    returned result will be set on the class in place of the factory method
    under the same name.

    ``@declared_attr`` is implemented differently than
    :class:`~sqlalchemy.ext.declarative.declared_attr` of SQLAlchemy, but they
    are both more often used on mixins to dynamically declare indices or
    constraints (also works for column and ``__table_args__``, or even normal
    class attributes)::

        class TrackedMixin:
            created = db.Column(db.DateTime(timezone=True))

            @db.declared_attr
            def unique_id(cls):
                return db.Column(db.Integer())

            @db.declared_attr
            def unique_constraint(cls):
                return db.UniqueConstraint('unique_id')

            @db.declared_attr
            def poly(cls):
                if cls.__name__ == 'Thing':
                    return db.Column(db.Unicode())

            @db.declared_attr
            def __table_args__(cls):
                if cls.__name__ == 'Thing':
                    return db.UniqueConstraint('poly'),

    .. note::

        This doesn't work if the model already had a ``__table__``.

    """
    m.__declared_attr__ = True
    return m


class Model:
    __metadata__ = None
    __table__ = None
    __attr_factory__ = ColumnAttribute

    def __init__(self):
        self.__values__ = {}

    @classmethod
    def _init_table(cls, sub_cls):
        table_name = None
        columns = []
        inspected_args = []
        updates = {}
        column_name_map = InvertDict()
        for each_cls in sub_cls.__mro__[::-1]:
            for k, v in getattr(each_cls, "__namespace__", each_cls.__dict__).items():
                declared_callable_attr = callable(v) and getattr(
                    v, "__declared_attr__", False
                )
                if k == "__tablename__":
                    if declared_callable_attr:
                        table_name = v(sub_cls)
                    else:
                        table_name = v
                    continue
                if declared_callable_attr:
                    v = updates[k] = v(sub_cls)
                if isinstance(v, sa.Column):
                    v = v.copy()
                    if not v.name:
                        v.name = k
                    column_name_map[k] = v.name
                    columns.append(v)
                    updates[k] = sub_cls.__attr_factory__(k, v)
                elif isinstance(v, (sa.Index, sa.Constraint)):
                    inspected_args.append(v)
        if table_name is None:
            return
        sub_cls._column_name_map = column_name_map

        # handle __table_args__
        table_args = updates.get(
            "__table_args__", getattr(sub_cls, "__table_args__", None)
        )
        args, table_kw = (), {}
        if isinstance(table_args, dict):
            table_kw = table_args
        elif isinstance(table_args, tuple) and table_args:
            if isinstance(table_args[-1], dict):
                args, table_kw = table_args[0:-1], table_args[-1]
            else:
                args = table_args

        args = (*columns, *inspected_args, *args)
        for item in args:
            try:
                _table = getattr(item, "table", None)
            except InvalidRequestError:
                _table = None
            if _table is not None:
                raise ValueError(
                    "{} is already attached to another table. Please do not "
                    "use the same item twice. A common mistake is defining "
                    "constraints and indices in a super class - we are working"
                    " on making it possible."
                )
        rv = sa.Table(table_name, sub_cls.__metadata__, *args, **table_kw)
        for k, v in updates.items():
            setattr(sub_cls, k, v)
        return rv


def declarative_base(metadata, model_classes=(Model,), name="Model"):
    return ModelType(name, model_classes, {"__metadata__": metadata})


# noinspection PyProtectedMember
@sa.inspection._inspects(ModelType)
def inspect_model_type(target):
    target._check_abstract()
    return sa.inspection.inspect(target.__table__)


__all__ = [
    "ColumnAttribute",
    "Model",
    "declarative_base",
    "declared_attr",
    "InvertDict",
]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gino', 'gino.dialects', 'gino.ext']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.2,<2.0', 'asyncpg>=0.18,<1.0']

extras_require = \
{':python_version < "3.7"': ['contextvars>=2.4,<3.0'],
 ':python_version < "3.8"': ['importlib_metadata>=1.3.0,<2.0.0']}

entry_points = \
{'sqlalchemy.dialects': ['asyncpg = gino.dialects.asyncpg:AsyncpgDialect',
                         'postgresql.asyncpg = '
                         'gino.dialects.asyncpg:AsyncpgDialect']}

setup_kwargs = {
    'name': 'gino',
    'version': '1.0.0rc2',
    'description': 'GINO Is Not ORM - a Python asyncio ORM on SQLAlchemy core.',
    'long_description': '====\nGINO\n====\n\n.. image:: https://img.shields.io/pypi/v/gino?logo=python&logoColor=white\n        :alt: PyPI Release Version\n        :target: https://pypi.python.org/pypi/gino\n\n.. image:: https://img.shields.io/github/workflow/status/python-gino/gino/test?label=test&logo=github\n        :alt: GitHub Workflow Status for tests\n        :target: https://github.com/python-gino/gino/actions?query=workflow%3Atest\n\n.. image:: https://img.shields.io/github/workflow/status/python-gino/gino/docs?label=docs&logo=github\n        :alt: GitHub Workflow Status for docs\n        :target: https://python-gino.org/docs/\n\n.. image:: https://img.shields.io/codacy/coverage/b6a59cdf5ca64eab9104928d4f9bbb97?logo=codacy\n        :alt: Codacy coverage\n        :target: https://app.codacy.com/gh/python-gino/gino/dashboard\n\n.. image:: https://img.shields.io/badge/Dependabot-active-brightgreen?logo=dependabot\n        :target: https://app.dependabot.com/accounts/python-gino/projects/129260\n        :alt: Dependabot\n\n.. image:: https://img.shields.io/gitter/room/python-gino/Lobby?logo=gitter\n        :target: https://gitter.im/python-gino/Lobby\n        :alt: Gitter chat\n\n\nGINO - GINO Is Not ORM - is a lightweight asynchronous ORM built on top of\nSQLAlchemy_ core for Python asyncio_. Now (early 2020) GINO supports only one\ndialect asyncpg_.\n\n* Free software: BSD license\n* Requires: Python 3.5\n\n\nDocumentation\n-------------\n\n* English_\n* Chinese_\n\n\nFeatures\n--------\n\n* Robust SQLAlchemy-asyncpg bi-translator with no hard hack\n* Asynchronous SQLAlchemy-alike engine and connection\n* Asynchronous dialect API\n* Asynchronous-friendly CRUD objective models\n* Well-considered contextual connection and transaction management\n* Reusing native SQLAlchemy core to build queries with grammar sugars\n* Support Sanic_, Tornado_, aiohttp_ and Quart_\n* Rich PostgreSQL JSONB support\n\n\nInstallation\n------------\n\n.. code-block:: console\n\n    pip install gino\n\n\nShowcase\n--------\n\n.. code-block:: python\n\n   import asyncio\n   from gino import Gino\n\n   db = Gino()\n\n\n   class User(db.Model):\n       __tablename__ = \'users\'\n\n       id = db.Column(db.Integer(), primary_key=True)\n       nickname = db.Column(db.Unicode(), default=\'noname\')\n\n\n   async def main():\n       await db.set_bind(\'postgresql://localhost/gino\')\n\n       # Create tables\n       await db.gino.create_all()\n\n       # Create object, `id` is assigned by database\n       u1 = await User.create(nickname=\'fantix\')\n       print(u1.id, u1.nickname)  # 1 fantix\n\n       # Returns all user objects with "d" in their nicknames\n       users = await User.query.where(User.nickname.contains(\'d\')).gino.all()\n       print(users)  # [<User object>, <User object>]\n\n       # Find one user object, None if not found\n       user = await User.query.where(User.nickname == \'daisy\').gino.first()\n       print(user)  # <User object> or None\n\n       # Execute complex statement and return command status\n       status, result = await User.update.values(\n           nickname=\'No.\' + db.cast(User.id, db.Unicode),\n       ).where(\n           User.id > 10,\n       ).gino.status()\n       print(status)  # UPDATE 8\n\n       # Iterate over the results of a large query in a transaction as required\n       async with db.transaction():\n           async for u in User.query.order_by(User.id).gino.iterate():\n               print(u.id, u.nickname)\n\n\n   asyncio.get_event_loop().run_until_complete(main())\n\n\nAbout The Name\n--------------\n\nAbout the name GINO Is Not ORM - because I don\'t really like ORM (smile). GINO\ndoes perform the Object-Relational Mapping work under the\n`Data Mapper Pattern`_, but it is just not a traditional ORM. The Objects in\nGINO are completely stateless from database - they are pure plain Python\nobjects in memory. Changing their attribute values does not make them "dirty" -\nor in a different way of thinking they are always "dirty". Any access to\ndatabase must be explicitly executed. Using GINO is more like making up SQL\nclauses with Models and Objects, executing them to make changes in database, or\nloading data from database and wrapping the results with Objects again. Objects\nare just row data containers, you are still dealing with SQL which is\nrepresented by Models and SQLAlchemy core grammars. Besides if you don\'t like\nORM at all, you can use GINO without ORM:\n\n.. code-block:: python\n\n    from gino import Gino\n\n    db = Gino()\n\n    user = db.Table(\n        \'users\', db,\n        db.Column(\'id\', db.BigInteger(), primary_key=True),\n        db.Column(\'nickname\', db.Unicode()),\n    )\n\n    async def main():\n        async with db.with_bind(\'postgresql://localhost/gino\'):\n            users = await db.select([user]).gino.all()\n            print(users)\n\n\n    import asyncio\n\n    asyncio.get_event_loop().run_until_complete(main())\n\n\nContribute\n----------\n\nThere are a few tasks in GitHub issues marked as ``help wanted``. Please feel\nfree to take any of them and pull requests are greatly welcome.\n\nTo run tests:\n\n.. code-block:: console\n\n   $ make install test\n\nFor the tests to be run, a database has to be available (please read more in CONTRIBUTING.rst)\n\nMeanwhile, these are also very much appreciated:\n\n* Just use it, tap the star and spread the love :)\n* Tell me about your story: gino@decentfox.com\n* Report_ a bug/issue, or request for new features\n* `Hire me`_ or `hire us`_ for your work (coding, consulting, troubleshooting)\n* `Become a patron`_\n\n\nProjects using GINO\n-------------------\n\n* AintQ_ - asyncio task queue on PostgreSQL\n* ExchangeratesAPI_ - Foreign exchange rates API with currency conversion\n\n\nCredits\n-------\n\nCredit goes to all contributors listed or not listed in the AUTHORS file. This\nproject is inspired by asyncpgsa_, peewee-async_ and asyncorm_. asyncpg_ and\nSQLAlchemy_ as the dependencies did most of the heavy lifting. This package was\ncreated with Cookiecutter_ and the `audreyr/cookiecutter-pypackage`_ project\ntemplate.\n\nSpecial thanks to my wife Daisy and her outsourcing company `DecentFoX Studio`_,\nfor offering me the opportunity to build this project. We are open for global\nsoftware project outsourcing on Python, iOS and Android development. And we are\nhiring_!\n\nGINO is developed proudly with |PyCharm|.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`audreyr/cookiecutter-pypackage`: https://github.com/audreyr/cookiecutter-pypackage\n.. _SQLAlchemy: https://www.sqlalchemy.org/\n.. _asyncpg: https://github.com/MagicStack/asyncpg\n.. _PostgreSQL: https://www.postgresql.org/\n.. _asyncio: https://docs.python.org/3/library/asyncio.html\n.. _Alembic: https://bitbucket.org/zzzeek/alembic\n.. _Sanic: https://github.com/channelcat/sanic\n.. _asyncpgsa: https://github.com/CanopyTax/asyncpgsa\n.. _peewee-async: https://github.com/05bit/peewee-async\n.. _asyncorm: https://github.com/monobot/asyncorm\n.. _Tornado: http://www.tornadoweb.org/\n.. _Quart: https://gitlab.com/pgjones/quart/\n.. _English: https://python-gino.org/docs/en/\n.. _Chinese: https://python-gino.org/docs/zh/\n.. _DecentFoX Studio: https://decentfox.com/\n.. _`Data Mapper Pattern`: https://en.wikipedia.org/wiki/Data_mapper_pattern\n.. _aiohttp: https://github.com/aio-libs/aiohttp\n.. _Report: https://github.com/python-gino/gino/issues\n.. _`Hire me`: https://www.linkedin.com/in/fantix/\n.. _`hire us`: https://decentfox.com/\n.. _`Become a patron`: https://www.patreon.com/fantixking\n.. _hiring: https://www.zhipin.com/gongsi/c6e283cf57f2d9361nF92NS7GA~~.html\n.. _contextvars: https://github.com/MagicStack/contextvars\n.. _replaced: https://github.com/MagicStack/contextvars/issues/2\n.. _`Python 3.7`: https://docs.python.org/3.7/library/contextvars.html\n.. _AintQ: https://github.com/fantix/aintq\n.. _ExchangeratesAPI: https://github.com/madisvain/exchangeratesapi\n.. |PyCharm| image:: ./docs/images/pycharm.svg\n        :height: 20px\n        :target: https://www.jetbrains.com/?from=GINO\n',
    'author': 'Fantix King',
    'author_email': 'fantix.king@gmail.com',
    'maintainer': 'Tony Wang',
    'maintainer_email': 'wwwjfy@gmail.com',
    'url': 'https://python-gino.org',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



import warnings
import types

from sqlalchemy import select
from sqlalchemy.schema import Column
from sqlalchemy.sql.elements import Label

from .declarative import Model


class Loader:
    @classmethod
    def get(cls, value):
        from .crud import Alias

        if isinstance(value, Loader):
            rv = value
        elif isinstance(value, type) and issubclass(value, Model):
            rv = ModelLoader(value)
        elif isinstance(value, Alias):
            rv = AliasLoader(value)
        elif isinstance(value, Column):
            rv = ColumnLoader(value)
        elif isinstance(value, Label):
            rv = ColumnLoader(value.name)
        elif isinstance(value, tuple):
            rv = TupleLoader(value)
        elif callable(value):
            rv = CallableLoader(value)
        else:
            rv = ValueLoader(value)
        return rv

    @property
    def query(self):
        rv = select(self.get_columns())
        from_clause = self.get_from()
        if from_clause is not None:
            rv = rv.select_from(from_clause)
        return rv.execution_options(loader=self)

    def do_load(self, row, context):
        raise NotImplementedError

    def get_columns(self):
        return []

    def get_from(self):
        return None

    def __getattr__(self, item):
        return getattr(self.query, item)


_none = object()


def _get_column(model, column_or_name) -> Column:
    if isinstance(column_or_name, str):
        return getattr(model, column_or_name)

    if isinstance(column_or_name, Column):
        if column_or_name in model:
            return column_or_name
        raise AttributeError(
            "Column {} does not belong to model {}".format(column_or_name, model)
        )

    raise TypeError(
        "Unknown column {} with type {}".format(column_or_name, type(column_or_name))
    )


class ModelLoader(Loader):
    def __init__(self, model, *columns, **extras):
        self.model = model
        self._distinct = None
        if columns:
            self.columns = [_get_column(model, name) for name in columns]
        else:
            self.columns = model
        self.extras = dict((key, self.get(value)) for key, value in extras.items())
        self.on_clause = None

    def _do_load(self, row):
        values = dict((c.name, row[c]) for c in self.columns if c in row)
        if all((v is None) for v in values.values()):
            return None
        rv = self.model()
        for c in self.columns:
            if c in row:
                # noinspection PyProtectedMember
                instance_key = self.model._column_name_map.invert_get(c.name)
                rv.__values__[instance_key] = row[c]
        return rv

    def do_load(self, row, context):
        distinct = True
        if self._distinct:
            if context is None:
                context = {}
            ctx = context.setdefault(self._distinct, {})
            key = tuple(row[col] for col in self._distinct)
            rv = ctx.get(key, _none)
            if rv is _none:
                rv = self._do_load(row)
                ctx[key] = rv
            else:
                distinct = False
        else:
            rv = self._do_load(row)

        if rv is None:
            return None, None
        else:
            for key, value in self.extras.items():
                value, distinct_ = value.do_load(row, context)
                if distinct_ is None:
                    continue

                if isinstance(getattr(self.model, key, None), types.FunctionType):
                    getattr(rv, key)(value)
                else:
                    setattr(rv, key, value)
            return rv, distinct

    def get_columns(self):
        yield from self.columns
        for subloader in self.extras.values():
            yield from subloader.get_columns()

    def get_from(self):
        rv = self.model
        for key, subloader in self.extras.items():
            from_clause = subloader.get_from()
            if from_clause is not None:
                rv = rv.outerjoin(from_clause, getattr(subloader, "on_clause", None))
        return rv

    def load(self, *columns, **extras):
        if columns:
            self.columns = [_get_column(self.model, name) for name in columns]

        self.extras.update((key, self.get(value)) for key, value in extras.items())
        return self

    def on(self, on_clause):
        self.on_clause = on_clause
        return self

    def distinct(self, *columns):
        self._distinct = columns
        return self

    def none_as_none(self, enabled=True):
        if not enabled:
            warnings.warn(
                "Disabling none_as_none is not supported.", DeprecationWarning,
            )
        return self


class AliasLoader(ModelLoader):
    def __init__(self, alias, *columns, **extras):
        super().__init__(alias, *columns, **extras)


class ColumnLoader(Loader):
    def __init__(self, column):
        self.column = column

    def do_load(self, row, context):
        return row[self.column], True


class TupleLoader(Loader):
    def __init__(self, values):
        self.loaders = tuple(self.get(value) for value in values)

    def do_load(self, row, context):
        return tuple(loader.do_load(row, context)[0] for loader in self.loaders), True


class CallableLoader(Loader):
    def __init__(self, func):
        self.func = func

    def do_load(self, row, context):
        return self.func(row, context), True


class ValueLoader(Loader):
    def __init__(self, value):
        self.value = value

    def do_load(self, row, context):
        return self.value, True
